@@ -2,9 +2,10 @@ import uuid
 import re
 import datetime
 import decimal
+import itertools
 
-from schematics.base import TypeException, NotAModelException
 
+from schematics.exceptions import ValidationError
 from schematics.types import schematic_types
 
 
@@ -18,45 +19,48 @@ class BaseTypeMetaClass(type):
 
 
 ###
-### Types
+### Base Type
 ###
 
 class BaseType(object):
-    """A base class for Types in a Structures model. Instances of this
+    """A base class for Types in a Schematics model. Instances of this
     class may be added to subclasses of `Model` to define a model schema.
     """
 
     __metaclass__ = BaseTypeMetaClass
 
-    def __init__(self, uniq_field=None, field_name=None, required=False,
-                 default=None, id_field=False, validation=None, choices=None,
-                 description=None, minimized_field_name=None):
+    def __init__(self, required=False, default=None, field_name=None,
+                 print_name=None, choices=None, validators=None, description=None,
+                 minimized_field_name=None):
 
-        self.uniq_field = '_id' if id_field else uniq_field or field_name
+        self.minimized_field_name = minimized_field_name
         self.field_name = field_name
+        self.print_name = print_name  ### TODO this name sucks
         self.required = required
         self.default = default
-        self.validation = validation
         self.choices = choices
-        self.id_field = id_field
+        self.validators = validators
         self.description = description
-        self.minimized_field_name = minimized_field_name
 
     def __get__(self, instance, owner):
         """Descriptor for retrieving a value from a field in a model. Do
-        any necessary conversion between Python and `Structures` types.
+        any necessary conversion between Python and `Schematics` types.
         """
         if instance is None:
             # Model class being used rather than a model object
             return self
 
+        if self.field_name not in instance._data:
+            raise AttributeError(self.field_name)
+    
         value = instance._data.get(self.field_name)
 
-        if value is None:
+        if value is None and self.default is not None:
             value = self.default
-            # Allow callable default values
+            # Callable values are best for mutable defaults
             if callable(value):
                 value = value()
+
         return value
 
     def __set__(self, instance, value):
@@ -64,39 +68,61 @@ class BaseType(object):
         """
         instance._data[self.field_name] = value
 
+    def __delete__(self, instance):
+        if self.name not in instance._fields:
+            instance_name = type(instance).__name__
+            error_msg = '%r has no attribute %r' % (instance_name, self.name)
+            raise AttributeError(error_msg)
+        del instance._fields[self.name]
+
+    ### TODO rename. `to_python`?
     def for_python(self, value):
-        """Convert a Structures type into native Python value
+        """Convert a Schematics type into native Python value
         """
         return value
 
+    ### TODO rename. `to_unicode`?
     def for_json(self, value):
-        """Convert a Structures type into a value safe for JSON encoding
+        """Convert a Schematics type into a value safe for JSON encoding
         """
         return self.for_python(value)
 
+    ###
+    ### Validation
+    ###
+
+    def validate_choices(self, value):
+        if self.choices is not None and value not in self.choices:
+            error_msg = 'Value must be one of %s.' % unicode(self.choices)
+            raise ValidationError(error_msg)
+        return value
+
+    def validate_required(self, value):
+        if self.required and value is None:
+            raise StopValidation(u'Field must have a significant value')
+        return value
+
     def validate(self, value):
-        """Perform validation on a value.
+        """Function that is overridden by subclasses for their validation logic
         """
         pass
 
     def _validate(self, value):
-        # check choices
-        if self.choices is not None:
-            if value not in self.choices:
-                raise TypeException("Value must be one of %s."
-                    % unicode(self.choices), self.field_name, value)
+        validator_chain = itertools.chain(
+            [
+                self.validate_required,
+                self.validate,
+                self.validate_choices,
+            ],
+            self.validators or []
+        )
 
-        # check validation argument
-        if self.validation is not None:
-            if callable(self.validation):
-                if not self.validation(value):
-                    raise TypeException('Value does not match custom'
-                                          'validation method.',
-                                           self.field_name, value)
-            else:
-                raise ValueError('validation argument must be a callable.')
+        for validator in validator_chain:
+            validator(value)
 
-        return self.validate(value)
+    ###
+    ### jsonschema
+    ###
 
     def _jsonschema_default(self):
         if callable(self.default):
@@ -141,6 +167,10 @@ class BaseType(object):
         return schema
 
 
+###
+### Standard Types
+###
+
 class UUIDType(BaseType):
     """A field that stores a valid UUID value and optionally auto-populates
     empty values with new UUIDs.
@@ -169,21 +199,65 @@ class UUIDType(BaseType):
         """Make sure the value is a valid uuid representation.  See
         http://docs.python.org/library/uuid.html for accepted formats.
         """
+        new_value = value
+        
         if not isinstance(value, (uuid.UUID,)):
             try:
-                value = uuid.UUID(value)
+                new_value = uuid.UUID(value)
             except ValueError:
-                raise TypeException('Not a valid UUID value',
-                    self.field_name, value)
-        return value
+                error_msg = 'Not a valid UUID value'
+                raise ValidationError(error_msg)
+
+        return True
 
     def for_json(self, value):
         """Return a JSON safe version of the UUID object.
         """
-
         return str(value)
 
 
+class IPv4Type(BaseType):
+    """ A field that stores a valid IPv4 address """
+
+    def __init__(self, auto_fill=False, **kwargs):
+        super(IPv4Type, self).__init__(**kwargs)
+
+    def _jsonschema_type(self):
+        return 'string'
+
+    @classmethod
+    def valid_ip(cls, addr):
+        try:
+            addr = addr.strip().split(".")
+        except AttributeError:
+            return False
+        try:
+            return len(addr) == 4 and all(int(octet) < 256 for octet in addr)
+        except ValueError:
+            return False
+
+    def validate(self, value):
+        """
+          Make sure the value is a IPv4 address:
+          http://stackoverflow.com/questions/9948833/validate-ip-address-from-list
+        """
+        if not IPv4Type.valid_ip(value):
+            error_msg = 'Invalid IPv4 address'
+            raise ValidationError(error_msg)
+        return True
+
+    def _jsonschema_format(self):
+        return 'ip-address'
+
+    @classmethod
+    def _from_jsonschema_formats(self):
+        return ['ip-address']
+
+    @classmethod
+    def _from_jsonschema_types(self):
+        return ['string']
+
+
 class StringType(BaseType):
     """A unicode string field.
     """
@@ -214,25 +288,24 @@ class StringType(BaseType):
     def _jsonschema_pattern(self):
         return self.regex
 
-    def for_python(self, value):
-        return unicode(value)
-
     def validate(self, value):
-        assert isinstance(value, (str, unicode))
+        value = unicode(value)
+        
+        if not isinstance(value, (str, unicode)):
+            raise ValidationError('Not a boolean')
 
         if self.max_length is not None and len(value) > self.max_length:
-            raise TypeException('String value is too long',
-                                  self.field_name, value)
+            raise ValidationError('String value is too long')
 
         if self.min_length is not None and len(value) < self.min_length:
-            raise TypeException('String value is too short',
-                                  self.uniq_field, value)
+            raise ValidationError('String value is too short')
 
         if self.regex is not None and self.regex.match(value) is None:
-            message = 'String value did not match validation regex',
-            raise TypeException(message, self.uniq_field, value)
+            error_msg = 'String value did not match validation regex'
+            raise ValidationError(error_msg)
 
         return value
+    
 
     def lookup_member(self, member_name):
         return None
@@ -271,7 +344,7 @@ class URLType(StringType):
 
     def validate(self, value):
         if not URLType.URL_REGEX.match(value):
-            raise TypeException('Invalid URL', self.field_name, value)
+            raise ValidationError('Invalid URL')
 
         if self.verify_exists:
             import urllib2
@@ -279,8 +352,7 @@ class URLType(StringType):
                 request = urllib2.Request(value)
                 urllib2.urlopen(request)
             except Exception:
-                message = 'URL does not exist'
-                raise TypeException(message, self.field_name, value)
+                raise ValidationError('URL does not exist')
 
         return value
 
@@ -302,8 +374,7 @@ class EmailType(StringType):
 
     def validate(self, value):
         if not EmailType.EMAIL_REGEX.match(value):
-            raise TypeException('Invalid email address', self.field_name,
-                                  value)
+            raise ValidationError('Invalid email address')
         return value
 
     def _jsonschema_format(self):
@@ -352,25 +423,21 @@ class NumberType(JsonNumberMixin, BaseType):
                 value = self.number_class(value)
         instance._data[self.field_name] = value    
 
-    def for_python(self, value):
-        return self.number_class(value)
-
     def validate(self, value):
         try:
             value = self.number_class(value)
         except:
-            raise TypeException('Not %s' % self.number_type, self.field_name,
-                                  value)
+            raise ValidationError('Not %s' % self.number_type)
 
         if self.min_value is not None and value < self.min_value:
-            raise TypeException('%s value below min_value: %s'
-                                  % (self.number_type, self.min_value),
-                                  self.field_name, value)
-
+            error_msg = '%s value below min_value: %s' % (self.number_type,
+                                                          self.min_value)
+            raise ValidationError(error_msg)
+        
         if self.max_value is not None and value > self.max_value:
-            raise TypeException('%s value above max_value: %s'
-                                  % (self.number_type, self.max_value),
-                                  self.field_name, value)
+            error_msg = '%s value above max_value: %s' % (self.number_type,
+                                                          self.max_value)
+            raise ValidationError(error_msg)
 
         return value
 
@@ -381,8 +448,8 @@ class IntType(NumberType):
 
     def __init__(self, *args, **kwargs):
         super(IntType, self).__init__(number_class=int,
-                                       number_type='Int',
-                                       *args, **kwargs)
+                                      number_type='Int',
+                                      *args, **kwargs)
 
     def _jsonschema_type(self):
         return 'number'
@@ -401,8 +468,8 @@ class LongType(NumberType):
     """
     def __init__(self, *args, **kwargs):
         super(LongType, self).__init__(number_class=long,
-                                        number_type='Long',
-                                        *args, **kwargs)
+                                       number_type='Long',
+                                       *args, **kwargs)
 
 
 class FloatType(NumberType):
@@ -410,8 +477,8 @@ class FloatType(NumberType):
     """
     def __init__(self, *args, **kwargs):
         super(FloatType, self).__init__(number_class=float,
-                                         number_type='Float',
-                                         *args, **kwargs)
+                                        number_type='Float',
+                                        *args, **kwargs)
 
 
 class DecimalType(BaseType, JsonNumberMixin):
@@ -422,31 +489,22 @@ class DecimalType(BaseType, JsonNumberMixin):
         self.min_value, self.max_value = min_value, max_value
         super(DecimalType, self).__init__(**kwargs)
 
-    def for_python(self, value):
-        if not isinstance(value, basestring):
-            value = unicode(value)
-        return decimal.Decimal(value)
-
-    def for_json(self, value):
-        return unicode(value)
-
     def validate(self, value):
         if not isinstance(value, decimal.Decimal):
             if not isinstance(value, basestring):
-                value = str(value)
+                value = unicode(value)
             try:
                 value = decimal.Decimal(value)
             except Exception:
-                raise TypeException('Could not convert to decimal',
-                                      self.field_name, value)
+                raise ValidationError('Could not convert to decimal')
 
         if self.min_value is not None and value < self.min_value:
-            raise TypeException('Decimal value below min_value: %s'
-                                  % self.min_value, self.field_name, value)
+            error_msg ='Decimal value below min_value: %s' % self.min_value
+            raise ValidationError(error_msg)
 
         if self.max_value is not None and value > self.max_value:
-            raise TypeException('Decimal value above max_value: %s'
-                                  % self.max_value, self.field_name, value)
+            error_msg = 'Decimal value above max_value: %s' % self.max_value
+            raise ValidationError(error_msg)
 
         return value
 
@@ -476,13 +534,11 @@ class MD5Type(BaseType, JsonHashMixin):
 
     def validate(self, value):
         if len(value) != MD5Type.hash_length:
-            raise TypeException('MD5 value is wrong length', self.field_name,
-                                  value)
+            raise ValidationError('MD5 value is wrong length')
         try:
-            int(value, 16)
+            value = int(value, 16)
         except:
-            raise TypeException('MD5 value is not hex', self.field_name,
-                                  value)
+            raise ValidationError('MD5 value is not hex')
         return value
 
 
@@ -493,13 +549,11 @@ class SHA1Type(BaseType, JsonHashMixin):
 
     def validate(self, value):
         if len(value) != SHA1Type.hash_length:
-            raise TypeException('SHA1 value is wrong length',
-                                  self.field_name, value)
+            raise ValidationError('SHA1 value is wrong length')
         try:
-            int(value, 16)
+            value = int(value, 16)
         except:
-            raise TypeException('SHA1 value is not hex', self.field_name,
-                                  value)
+            raise ValidationError('SHA1 value is not hex')
         return value
 
 
@@ -511,6 +565,9 @@ class BooleanType(BaseType):
     """A boolean field type.
     """
 
+    TRUE = ('True', 'true', '1')
+    FALSE = ('False', 'false', '0')
+
     def _jsonschema_type(self):
         return 'boolean'
 
@@ -522,12 +579,21 @@ class BooleanType(BaseType):
     def _from_jsonschema_formats(self):
         return [None]
 
-    def for_python(self, value):
-        return bool(value)
+    def __set__(self, instance, value):
+        """
+        Accept some form of True/False as string
+        """
+        if isinstance(value, (str, unicode)):
+            if value in BooleanType.TRUE:
+                value = True
+            elif value in BooleanType.FALSE:
+                value = False
+
+        instance._data[self.field_name] = value
 
     def validate(self, value):
         if not isinstance(value, bool):
-            raise TypeException('Not a boolean', self.field_name, value)
+            raise ValidationError('Not a boolean')
         return value
 
 
@@ -538,8 +604,16 @@ class DateTimeType(BaseType):
     def _jsonschema_type(self):
         return 'string'
 
-    def __init__(self, format=lambda dt: dt.isoformat(), **kwargs):
-        self.format = format
+    def __init__(self, format=None, **kwargs):
+        if format is None:
+            def formatter(dt):
+                if dt is None:
+                    return None
+                else:
+                    return dt.isoformat()
+            self.format = formatter
+        else:
+            self.format = format
         super(DateTimeType, self).__init__(**kwargs)
 
     def _jsonschema_format(self):
@@ -585,6 +659,11 @@ class DateTimeType(BaseType):
         iso8601 = '(\d\d\d\d)-(\d\d)-(\d\d)' \
                   'T(\d\d):(\d\d):(\d\d)(?:\.(\d\d\d\d\d\d))?'
         elements = re.findall(iso8601, datestring)
+        
+        if len(elements) < 1:
+            error_msg = 'Date string could not transform to datetime'
+            raise ValidationError(error_msg)
+        
         date_info = elements[0]
         date_digits = [int(d) for d in date_info if d]
         value = datetime.datetime(*date_digits)
@@ -602,20 +681,17 @@ class DateTimeType(BaseType):
         elif hasattr(format, '__call__'):
             iso_dt = format(dt)
         else:
-            raise TypeException('DateTimeType format must be a string or callable')
+            error_msg = 'DateTimeType format must be a string or callable'
+            raise ValidationError(error_msg)
         return iso_dt
 
     def validate(self, value):
         if not isinstance(value, datetime.datetime):
-            raise TypeException('Not a datetime', self.field_name, value)
-        return value
-
-    def for_python(self, value):
+            raise ValidationError('Not a datetime')
         return value
 
     def for_json(self, value):
-        v = DateTimeType.date_to_iso8601(value, self.format)
-        return v
+        return DateTimeType.date_to_iso8601(value, self.format)
 
 
 class DictType(BaseType):
@@ -628,8 +704,11 @@ class DictType(BaseType):
 
     def __init__(self, basecls=None, *args, **kwargs):
         self.basecls = basecls or BaseType
+        
         if not issubclass(self.basecls, BaseType):
-            raise NotATypeException('basecls is not subclass of BaseType')
+            error_msg = 'basecls is not subclass of BaseType'
+            return ValidationError(error_msg)
+
         kwargs.setdefault('default', lambda: {})
         super(DictType, self).__init__(*args, **kwargs)
 
@@ -637,17 +716,17 @@ class DictType(BaseType):
         """Make sure that a list of valid fields is being used.
         """
         if not isinstance(value, dict):
-            raise TypeException('Only dictionaries may be used in a '
-                                  'DictType', self.field_name, value)
+            error_msg = 'Only dictionaries may be used in a DictType'
+            raise ValidationError(error_msg)
 
+        ### TODO this can probably be removed
         if any(('.' in k or '$' in k) for k in value):
-            raise TypeException('Invalid dictionary key name - keys may not '
-                                  'contain "." or "$" characters',
-                                  self.field_name, value)
+            error_msg = 'Invalid dictionary key - may not contain "." or "$"'
+            raise ValidationError(error_msg)
         return value
 
     def lookup_member(self, member_name):
-        return self.basecls(uniq_field=member_name)
+        return self.basecls(field_name=member_name)
 
 
 class GeoPointType(BaseType):
@@ -670,21 +749,20 @@ class GeoPointType(BaseType):
         """Make sure that a geo-value is of type (x, y)
         """
         if not len(value) == 2:
-            raise TypeException('Value must be a two-dimensional point',
-                                  self.field_name, value)
+            error_msg = 'Value must be a two-dimensional point'
+            raise ValidationError(error_msg)
         if isinstance(value, dict):
             for v in value.values():
                 if not isinstance(v, (float, int)):
                     error_msg = 'Both values in point must be float or int'
-                    raise TypeException(error_msg, self.field_name, value)
+                    raise ValidationError(error_msg)
         elif isinstance(value, (list, tuple)):
             if (not isinstance(value[0], (float, int)) and
                 not isinstance(value[1], (float, int))):
                 error_msg = 'Both values in point must be float or int'
-                raise TypeException(error_msg, self.field_name, value)
+                raise ValidationError(error_msg)
         else:
-            raise TypeException('GeoPointType can only accept tuples, '
-                                  'lists of (x, y), or dicts of {k1: v1, '
-                                  'k2: v2}',
-                                  self.field_name, value)
+            error_msg = 'GeoPointType can only accept tuples, lists, or dicts'
+            raise ValidationError(error_msg)
+        
         return value
