@@ -5,9 +5,11 @@ except:
 from operator import itemgetter
 
 from schematics.models import Model
-from schematics.base import  TypeException
 from schematics.types import BaseType, DictType
 from schematics.datastructures import MultiValueDict
+from schematics.serialize import to_python, to_json, for_jsonschema
+from schematics.validation import validate
+from schematics.exceptions import ValidationError
 
 
 RECURSIVE_REFERENCE_CONSTANT = 'self'
@@ -19,26 +21,32 @@ class ListType(BaseType):
     """
 
     def __init__(self, fields, **kwargs):
-        # Some helpful functions
-        is_basetype = lambda tipe: isinstance(tipe, BaseType)
-        is_model = lambda tipe: isinstance(tipe, ModelType)
-        is_dicttype = lambda tipe: isinstance(tipe, DictType)
+        super(ListType, self).__init__(**kwargs)
 
-        # field instance
+        ### Short hand
+        is_basetype = lambda field: isinstance(field, BaseType)
+        is_model = lambda field: isinstance(field, ModelType)
+        is_dicttype = lambda field: isinstance(field, DictType)
+
+        ### fields is a schematic Type
         if is_basetype(fields):
             if is_model(fields):
                 kwargs.setdefault('primary_embedded', fields)
             fields = [fields]
-        # something other than a list
+            
+        ### something other than a list
         elif not isinstance(fields, list):
-            raise TypeException('Argument to ListType constructor must be '
-                                'a valid field or list of fields',
-                                self.field_name, list)
-        # some bad stuff in the list
+            error_msg = 'Argument to ListType constructor must be '
+            error_msg = error_msg + 'a valid field or list of fields'
+            
+            raise ValidationError(error_msg)
+        
+        ### some bad stuff in the list
         elif list(ifilterfalse(is_basetype, fields)):
-            raise TypeException('Argument to ListType constructor must be '
-                                'a valid field or list of valid fields',
-                                self.field_name, list)
+            error_msg = 'Argument to ListType constructor must be '
+            error_msg = error_msg + 'a valid field or list of valid fields'
+            raise ValidationError(error_msg)
+        
         else:
             models = filter(is_model, fields)
             dicts = filter(is_dicttype, fields)
@@ -46,35 +54,64 @@ class ListType(BaseType):
                 kwargs.setdefault('primary_embedded', None)
             if models:
                 kwargs.setdefault('primary_embedded', models[0])
+
         self.fields = fields
         kwargs.setdefault('default', list)
 
         self.primary_embedded = kwargs.pop('primary_embedded', None)
-        super(ListType, self).__init__(**kwargs)
 
-    def __set__(self, instance, value):
+    def __set__(self, instance, value_list):
         """Descriptor for assigning a value to a type in a model.
         """
+        new_value = value_list
+
         is_model = lambda tipe: isinstance(tipe, ModelType)
         model_fields = filter(is_model, self.fields)
+
         if self.primary_embedded:
             model_fields.remove(self.primary_embedded)
             model_fields.insert(0, self.primary_embedded)
 
-        if value is None:
-            value = []  # have to use a list
+        if value_list is None:
+            value_list = []  # have to use a list
+
+        errors_found = False
 
         if model_fields:
-            list_of_models = list()
-            for model in value:
-                if isinstance(model, dict):
-                    for model_field in model_fields:
-                        model_obj = model_field.model_type_obj(**model)
-                        model_obj.validate()
-                        model = model_obj
-                list_of_models.append(model)
-            value = list_of_models
-        instance._data[self.field_name] = value
+            new_data = list()
+            for datum in value_list:
+                datum_instance = None
+                is_dict = False
+
+                ### Extract field names from datum
+                datum_fields = None
+                if isinstance(datum, dict):
+                    datum_fields = datum.keys()
+                else:
+                    datum_fields = datum._fields.keys()
+                    
+                ### Determine matching model
+                for model_field in model_fields:
+                    test_keys = model_field.model_type_obj._fields.keys()
+
+                    if len(set(datum_fields) - set(test_keys)) == 0:
+                        if datum is None:
+                            datum = {}
+                        elif not isinstance(datum, dict):
+                            datum = datum._data
+                        datum_instance = model_field.model_type_obj(**datum)
+
+                ### Validate model
+                try:
+                    result = datum_instance.validate()
+                    new_data.append(datum_instance)
+                except ValidationError, ve:
+                    errors_found = True
+                    
+            new_value = new_data
+
+        if not errors_found:
+            instance._data[self.field_name] = new_value
 
     def _jsonschema_type(self):
         return 'array'
@@ -83,6 +120,7 @@ class ListType(BaseType):
     def _from_jsonschema_types(self):
         return ['array']
 
+
     @classmethod
     def _from_jsonschema_formats(self):
         return [None]
@@ -94,7 +132,8 @@ class ListType(BaseType):
         for item in value:
             for field in self.fields:
                 try:
-                    yield getattr(field, output_format_method_name)(item)
+                    value = getattr(field, output_format_method_name)(item)
+                    yield value
                 except ValueError:
                     continue
 
@@ -112,19 +151,26 @@ class ListType(BaseType):
         """
         if not isinstance(value, (list, tuple)):
             error_msg = 'Only lists and tuples may be used in a list field'
-            raise TypeException(error_msg, self.field_name, value)
+            raise ValidationError(error_msg)
 
-        if not self.fields:  # if we want everything to validate
-            return
+        if not self.fields:
+            return value
 
+        errors = []
+        good_data = []
         for item in value:
-            try:
-                for field in self.fields:
+            for field in self.fields:
+                try:
                     field.validate(item)
-            except Exception, e:
-                raise TypeException('Invalid ListType item',
-                                      self.field_name, str(item))
-        return value
+                    value = field.for_python(item)
+                    good_data.append(value)
+                except ValidationError, ve:
+                    errors.append(ve)
+
+        if len(errors) > 0:
+            raise ValidationError('Invalid ListType item')
+        
+        return good_data
 
     def _set_owner_model(self, owner_model):
         for field in self.fields:
@@ -182,16 +228,13 @@ class ModelType(BaseType):
         is_embeddable = lambda dt: issubclass(dt, Model)
         if not isinstance(model_type, basestring):
             if not model_type or not is_embeddable(model_type):
-                raise TypeException('Invalid model class provided to an '
-                                    'ModelType',
-                                    self.field_name, model_type)
+                error_msg = 'Invalid model class provided to an ModelType'
+                raise ValidationError(error_msg)
         self.model_type_obj = model_type
         super(ModelType, self).__init__(**kwargs)
 
     def __set__(self, instance, value):
-        if value is None:
-            return
-        if not isinstance(value, self.model_type):
+        if value != None and not isinstance(value, self.model_type):
             value = self.model_type(**value)
         instance._data[self.field_name] = value
 
@@ -216,13 +259,13 @@ class ModelType(BaseType):
         return [None]
 
     def for_jsonschema(self):
-        return self.model_type.for_jsonschema()
+        return for_jsonschema(self.model_type)
 
     def for_python(self, value):
-        return value.to_python()
+        return to_python(value)
 
     def for_json(self, value):
-        return value.to_json(encode=False)
+        return to_json(value, encode=False)
 
     def validate(self, value):
         """Make sure that the model instance is an instance of the
@@ -230,11 +273,9 @@ class ModelType(BaseType):
         """
         # Using isinstance also works for subclasses of self.model
         if not isinstance(value, self.model_type):
-            raise TypeException('Invalid modeltype instance '
-                                  'provided to an ModelType',
-                                  self.field_name, value)
-        self.model_type.validate(value)
-        return value
+            error_msg = 'Invalid modeltype instance provided to an ModelType'
+            raise ValidationError(error_msg)
+        return value.validate()
 
     def lookup_member(self, member_name):
         return self.model_type._fields.get(member_name)
@@ -243,8 +284,11 @@ class ModelType(BaseType):
 class MultiValueDictType(DictType):
     def __init__(self, basecls=None, *args, **kwargs):
         self.basecls = basecls or BaseType
+        
         if not issubclass(self.basecls, BaseType):
-            raise NotAModelException('basecls is not subclass of BaseType')
+            error_msg = 'basecls is not subclass of BaseType'
+            raise ValidationError(error_msg)
+        
         kwargs.setdefault('default', lambda: MultiValueDict())
         super(MultiValueDictType, self).__init__(*args, **kwargs)
 
@@ -258,14 +302,15 @@ class MultiValueDictType(DictType):
         """Make sure that a list of valid fields is being used.
         """
         if not isinstance(value, (dict, MultiValueDict)):
-            raise TypeException('Only dictionaries or MultiValueDict may be '
-                                  'used in a DictType', self.field_name,
-                                  value)
+            error_msg = 'Only dictionaries or MultiValueDict may be used in a '
+            error_msg = error_msg + 'DictType'
+            raise ValidationError(error_msg)
 
+        ### TODO get rid of this
         if any(('.' in k or '$' in k) for k in value):
-            raise TypeException('Invalid dictionary key name - keys may not '
-                                  'contain "." or "$" characters',
-                                  self.field_name, value)
+            error_msg = 'Invalid dictionary key name - keys may not contain '
+            error_msg = error_msg + '"." or "$" characters'
+            raise ValidationError(error_msg)
         return value
 
     def for_json(self, value):
